SystemOrganization addCategory: #AmbientTalkST!
SystemOrganization addCategory: #'AmbientTalkST-Benchmarks'!

Message subclass: #ATMessage
	instanceVariableNames: 'sender receiver result'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST'!

!ATMessage class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 17:06'!
from: aMessage
	"Used to convert a normal Message object to an ATMessage object"
	| newATMsg |
	newATMsg := self new.
	newATMsg setSelector: aMessage selector arguments: aMessage arguments.
	newATMsg lookupClass: aMessage lookupClass.
	^ newATMsg.! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 15:11'!
initialize
	super initialize.
	result := ATPromise new.! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:20'!
promise
	^ result! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
receiver
	^ receiver! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
receiver: anObj
	receiver := anObj! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 1/1/2012 21:48'!
result
	"Result is blocking on the promise.
	 Use #promise to obtain the promise directly."
	^ result read! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:20'!
result: anObj
	result deliver: anObj! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 20:38'!
send
	"answer the result of sending this message to its receiver"

	^ receiver perform: selector withArguments: args! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
sender
	^ sender! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
sender: anObj
	sender := anObj! !

BenchmarkGameSuite subclass: #ATBenchmarkGameSuite
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Benchmarks'!

!ATBenchmarkGameSuite methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 17:39'!
benchATBinaryTrees
	| bench problemSize treeActor benchActor treeFactory result |
	problemSize := self problemSize
			ifNotNil: [ self problemSize ]
			ifNil:    [ BGBinaryTrees defaultProblemSize ].
	
	treeActor := ATActor new.
	benchActor := ATActor new.
	
	treeFactory := ATFarReference for: BGTreeNode in: treeActor. 
	
	bench := BGBinaryTrees new.
	bench binaryTreeFactory: treeFactory.
	
	result := benchActor evaluateAsync: [
		bench binarytrees: problemSize to: (TextStream with: String new).
	].
	
	^ result read.! !

!ATBenchmarkGameSuite methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 22:41'!
benchATNBody
	| problemSize bodiesActor bodies |
	problemSize := self problemSize
			ifNotNil: [ self problemSize ]
			ifNil:    [ BGNBody defaultProblemSize ].

	bodiesActor := ATActor new.
	
	bodies := ATFarReference for: BGNBodySystem new in: bodiesActor. 
	
	^ BGNBody nbody: problemSize with: bodies.
! !

TestCase subclass: #ATBenchmarkGameSuiteTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Benchmarks'!

!ATBenchmarkGameSuiteTests methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 22:52'!
testATBinaryTrees
	| bench |
	bench := ATBenchmarkGameSuite new.
	self assert: -1 equals: bench benchATBinaryTrees.! !

!ATBenchmarkGameSuiteTests methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 22:53'!
testATNBody
	| bench result |
	bench := ATBenchmarkGameSuite new.
	result := bench benchATNBody.
	
	self assert: (((result at: 2) asString) beginsWith: '-0.169087605')! !

Object subclass: #ATActor
	instanceVariableNames: 'mailbox terminate process'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST'!

!ATActor methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 17:08'!
enqueue: aMessage
	"Directly put a message into the actors mailbox."
	mailbox nextPut: aMessage! !

!ATActor methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 15:09'!
evaluateAsync: aBlock
	"This is an implementation level method.
	 The block is by convention not referring to
	 unsafe shared state."
	| msg |
	msg := ATMessage new.
	msg receiver: aBlock.
	msg setSelector: #value arguments: #().

	mailbox nextPut: msg.
	
	^ msg promise.! !

!ATActor methodsFor: 'initialize-release' stamp: 'StefanMarr 11/25/2011 17:00'!
initialize
	super initialize.
	mailbox := SharedQueue new.
	
	terminate := false.
	
	process := [[terminate] whileFalse: [ self processIncomingMessages. ]] fork.! !

!ATActor methodsFor: 'initialize-release' stamp: 'StefanMarr 1/1/2012 20:52'!
processIncomingMessages
	| msg result |
	msg := mailbox next.
	result := msg send.
	
	msg result: result! !

Object subclass: #ATFarReferenceBase
	instanceVariableNames: 'obj actor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST'!

!ATFarReferenceBase commentStamp: 'StefanMarr 1/1/2012 16:45' prior: 0!
ATFarReferenceBase is providing the data and #cannotInterpret: for ATFarReference.
Since #cannotInterpret: is always send to the super class, we needed this base class.
Refer to ATFarReference for relevant class-side methods.

Instance Variables:
	obj	   - the object this far-reference is pointing to
	actor  - the actor the referenced object is contained in!

ATFarReferenceBase subclass: #ATFarReference
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST'!

!ATFarReference commentStamp: 'StefanMarr 1/1/2012 16:51' prior: 0!
ATFarReference implements a far-reference, i.e., a proxy for an object that is referenced from another actor.
The base class ATFarReferenceBase provides the proxy implementation.
The methodDict of ATFarReference is set to nil to be able to use the #cannotInterpet: mechanism, which enables us
to intercept all messages sent to an object, including #doesNotUnderstand.

See the class-side methods to instantiate a far-reference.!

!ATFarReference class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 21:45'!
deStratify
	self methodDict: self emptyMethodDictionary.! !

!ATFarReference class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 23:04'!
for: anObj in: anActor
	| ref |
	"SmallIntegers and bools are immutable and directly used."
	((anObj isKindOf: SmallInteger)
	  or: [anObj isKindOf: Boolean])
		ifTrue: [^anObj].
	
	"Strings are copied as by-value objects kind of isolates"
	(anObj isKindOf: String)
		ifTrue: [^ anObj copy].
		
	ref := ATFarReferenceBase new.
	ref object: anObj.
	ref actor:  anActor.
	
	ATFarReference adoptInstance: ref.
	
	^ ref! !

!ATFarReference class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 16:19'!
initialize
	"self initialize."
	self methodDict: nil.! !

!ATFarReference class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 22:06'!
isStratified
	^ methodDict isNil! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 16:22'!
actor
	^ actor! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 16:22'!
actor: anActor
	actor := anActor! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 22:10'!
cannotInterpret: aMessage
	| msg |
	Transcript cr; show: '#ci: rcvr: ', obj asString, ' msg: ', aMessage asString.
	
	msg := ATMessage from: aMessage.
	msg receiver: obj.
	actor enqueue: msg.
	
	^ ATFarReference for: msg result in: actor.! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 21:11'!
doesNotUnderstand: aMessage 
	"Try this for debugging"
	^ self cannotInterpret: aMessage! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 16:22'!
object
	^ obj! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 16:22'!
object: anObj
	obj := anObj! !

Object subclass: #ATIsolate
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST'!

Object subclass: #ATObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST'!

Object subclass: #ATPromise
	instanceVariableNames: 'signal value'
	classVariableNames: 'NoValueDelivered'
	poolDictionaries: ''
	category: 'AmbientTalkST'!

!ATPromise class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 14:20'!
example
	"This is not a real test case, but just to see how it is supposed to work"
	| p |
	p := ATPromise new.
	
	[|val| 
	 Transcript log: 'About to read promised value.'.
	 val := p read.
	 Transcript log: 'Val: ', val asString.] fork.

	Processor yield.
	
	Transcript log: 'About to deliver promised value.'.
	p deliver: 42.
	Transcript log: 'Deliver promised value.'.
	! !

!ATPromise class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/4/2011 15:53'!
initialize
	NoValueDelivered := Object new.! !

!ATPromise methodsFor: 'announcing' stamp: 'StefanMarr 11/4/2011 16:07'!
deliver: aValue
	"Deliver the promised value."
	
	
	"Warning: this is racy, but we don't care for the moment.
	 Should be a single produce anyway. Everything else is a bug.
	 And, we do at least best effort error notification..."
	value = NoValueDelivered ifFalse: [
		| err |
		err := Error new.
		err messageText: 'Promise has been delivered more than once.'.
		err signal.
	].

	value := aValue.
	
	signal signal.
	
	^ value.! !

!ATPromise methodsFor: 'initialization' stamp: 'StefanMarr 11/4/2011 16:09'!
initialize
	value := NoValueDelivered.
	signal := Semaphore new.! !

!ATPromise methodsFor: 'accessing' stamp: 'StefanMarr 11/4/2011 16:01'!
read
	[value = NoValueDelivered] whileTrue: [
		"This is ugly, but I do not see how I can prevent the data
		 race for signals otherwise.
		 In case there are multiple possible consumers everyone
		 but the first will starve when they arrive before the
		 promise was delivered.
		 Now, we just have a delay, but it is guaranteed to return."
		signal waitTimeoutMSecs: 100.
	].
	^ value! !

SMarkLoops subclass: #ATBenchLoops
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Benchmarks'!

!ATBenchLoops methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 15:00'!
benchATFloatLoop
	| promise actor |
	
	actor := ATActor new.
	promise := actor evaluateAsync: [
		self benchFloatLoop.
		#done
	].

	[promise read = #done] assert.! !

!ATBenchLoops methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 15:57'!
benchATIntLoop
	| promise actor |
	
	actor := ATActor new.
	promise := actor evaluateAsync: [
		self benchIntLoop.
		#done
	].

	[promise read = #done] assert.! !
ATFarReference initialize!
ATPromise initialize!
